% Created by Julia Baum on September 19, 2011
%
% This code splits the day of data into 15 minute bins (900 seconds) and
% then calls the optimization function DepMILP to perform the optimization.
% If a flight has an optimized takeoff time that is no longer in the
% current bin then move it to the next bin and optimize the next bin.
%
% INPUTS: 
% depData       A days worth of departure data from DFW
%
% depData consists of the spot ready times (real data), unimpeded times
% (derived fuzzy times), runway ready times (spot ready + unimpeded times,
% refered to as alpha), and the aircraft types (1-4) for a full day of data
% from DFW.
%
% Change binSize to determine how large the bins should be.
%
% OUTPUTS: 
% A             Aircraft types 
% B             Spot leave times
% D             Unimpeded taxi times
% E             Runway ready times (alphas) 
% F             Optimized takeoff times (t's) 
% G             Optimized runway sequence 
% H             Original spot sequence 
% I             Optimized spot sequence 
% J             Optimized spot leave times (calculated by t-unimpeded times)
% detFCFSt      First come first serve (FCFS) takeoff times for deterministic data 
% delayFCFS     System delay associated with deterministic FCFS takeoff times (calculated by detFCFSt-alphas)
% optDelay      System delay associated with deterministic optimization
%               takeoff times (calculated by t-alphas)
% solution      This is the output of the optimization and is useful for
%               determining whether there was a problem during the
%               optimization
% yalmipTime    The amount of time it took yalmip to set up the MILP
%               optimization
% solverTime    The amount of time it took to solve the MILP optimization

function [A, B, D, E, F, G, H, I, J, detFCFSt, delayFCFS, optDelay, solution, yalmipTime, solverTime] = DepBinning(depData)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Initialize variables %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

binSize = 900; % Determines the size of the bins in seconds. 900 seconds = 15 minutes
minData = min(depData);
aircraft1 = minData(3); % Start time for the first bin
maxData = max(depData);
aircraftN = maxData(3); % End time for the last bin
depDataChunk = []; % A chunk of data in the bin's runway ready time period that is going to be optimized
spotPlaceChunk = []; % A chunk of the spot sequence that goes with the data chunk
numDep = []; % Number of flights in a bin
lastFlight = 0; % Take off time of the last flight in the previous bin
lastPlace = 0; % The position of the last flight to takeoff in the previous bin
lastAtype = []; % The aircraft type of the last flight to takeoff in the previous bin
binNum = 1; % The number of the bin that is currently being solved for
optDelay = 0; % The system delay for the optimal take off times
A = []; % Aircraft types
B = []; % Spot leave times
D = []; % Original unimpeded taxi times
E = []; % Original simulated runway ready times
F = []; % Optimal take off times
G = []; % Optimal sequence at the runway
H = []; % Original sequence at the spot
I = []; % Optimal sequence at the spot
J = []; % Optimal spot leave times
yalmipTime = 0; % Yalmip's total setup time
solverTime = 0; % Total solving time
alphaSortData = sortrows(depData,3); % Sort the data by runway ready times (alphas)
alphaSort = alphaSortData(:,3); % Sorted alphas
aTypeSort = alphaSortData(:,4); % Sorted aircraft types

% Separation time matrix: leading rows, trailing columns
separationT = [60, 60, 60, 60;
               90, 60, 60, 60;
               120, 120, 90, 120;
               120, 90, 90, 90];

% NASA's separation matrix: leading rows, trailing columns (opposite their
% paper) separationT = [59, 59, 59, 59;
%                       88, 61, 61, 61; 
%                       109, 109, 90, 109; 
%                       110, 91, 91, 91];

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% FCFS at runway %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

[detFCFSt, alphaSort, deltaSort] = newRunwayTimes(alphaSortData, separationT);

% System delay for FCFS
delayFCFS = sum(detFCFSt - alphaSort);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Separating the day into 15 minute bins %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Sort the aircrafts into binSize seconds windows, run the
% optimization on each window keeping the correct amoung of separation
% times between bins.

% Find the original sequence at the spot based on original spot leave times
    spotPlace = [];
    [~, spotPlacePos] = sort(alphaSortData(:,1)); % Position of ready at spot sequence
    for i = 1 : size(alphaSortData,1)
        spotPlace(i,1) = find(spotPlacePos(:,1) == i); % Original sequence at the spot
    end

while aircraft1 <= aircraftN % Go through all of the runway ready times in the data      
   % Check if the aircraft is ready during this time window
    for i = 1 : size(alphaSortData, 1)
        if alphaSortData(i,3) >= aircraft1 && alphaSortData(i,3) < aircraft1 + binSize; 
            % if so put it in a chunk of data along with other aircraft in this bin
            depDataChunk = [depDataChunk; alphaSortData(i,:)]; 
            spotPlaceChunk = [spotPlaceChunk; spotPlace(i)];        
        end
    end
    spotPlaceChunk = spotPlaceChunk - lastPlace;
    if size(depDataChunk,1) > 0
        % Check if there are any aircraft who are ready at the spot before 
        % any of the aircraft already in the bin who are ready at the runway 
        % after the bin's upper boundary
        maxChunk = max(depDataChunk);
        numFlights = size(depDataChunk,1);
        
        [t, place, solution, So, spotPlaceOpt, deltaChunk] = DepMILP(depDataChunk, numFlights, lastFlight, lastAtype, binNum, separationT, spotPlaceChunk);
        if solution.problem ~= 0
            return
        end
        
        spots = depDataChunk(:,1);          % Original ready to leave spot times
        unimps = depDataChunk(:,2);         % Unimpeded times 
        alphas = depDataChunk(:,3);         % Ready at runway times 
        aTypes = depDataChunk(:,4);         % aircraft types
        t = double(t);                      % Assigned take off times 
        So = double(So);                    % Optimal spot leave times
        spotPlaceOpt = double(spotPlaceOpt);% Optimal spot sequence
        place = double(place);              % Optimal takeoff sequence 
        
        totalDelay = sum(t-alphas,1); % System delay
        yalmipTime = yalmipTime + solution.yalmiptime; % Total yalmip time
        solverTime = solverTime + solution.solvertime; % Total solver time
        
        % Sequences begin where last bin left off
        spotPlaceChunk = spotPlaceChunk + lastPlace;
        spotPlaceOpt = spotPlaceOpt + lastPlace;
        place = place + lastPlace;
        
        % Store data from bin in a matrix with the rest of the bins' data
        A = [A; aTypes]; % Aircraft types
        B = [B; spots]; % Spot leave times
        D = [D; unimps]; % Original unimpeded taxi times
        E = [E; alphas]; % Original simulated runway ready times
        F = [F; t]; % Optimal take off times
        G = [G; place]; % Optimal sequence at the runway
        H = [H; spotPlaceChunk]; % Original sequence at the spot
        I = [I; spotPlaceOpt]; % Optimal sequence at the spot
        J = [J; So]; % Optimal spot leave times
        
        % Store information of last flight, so that the next bin obeys
        % separation constraints
        lastFlight = max(t);
        lastPlace = max(G);
        for i = 1 : size(depDataChunk)
            if t(i) == lastFlight
                lastAtype = depDataChunk(i,4);
            end
        end
        numDep = [numDep; numFlights]; % Keeps track of how many aircraft are in each bin
        depDataChunk = [];
        spotPlaceChunk = [];
    else
        numFlights = size(depDataChunk,1); % Number of flights in the bin
        totalDelay = 0;
        numDep = [numDep; numFlights]; % Keeps track of how many aircraft are in each bin
    end
    % Prepare for next bin
    aircraft1 = aircraft1 + binSize; % Increase bin boundary time
    binNum = binNum + 1; % Increase bin number
    optDelay = optDelay + totalDelay; % Add the system delay of this bin to the total system delay
end


