% Created by Julia Baum on September 19, 2011
% 
% This code uses yalmip to set up a Mixed Integer Linear Programming (MILP)
% optimization and then solves it with Gurobi. The objective function is to
% minimize system delay. It could be changed to minimize makespan or
% minimize the maximum delay. Change numCPS to restrict position shifting
% or disclude constraint 8 to remove constrained position shifting (CPS).
% 
% INPUTS: 
% depDataChunk      A chunk of the data whose alpha runway ready times are in
%                   the current bin
% numFlights        The number of flights in the current bin
% lastFlight        The takeoff time of the last flight in the previous bin 
% lastAtype         The aircraft type of the last flight in the previous bin
% binNum            The number bin that is currently being optimized
% separationT       A matrix of separation times dependent on aircraft types
% spotPlaceChunk    Position of ready at spot sequence
%
% OUTPUTS:
% t                 Optimized takeoff times
% place             Optimized takeoff runway sequence
% solution          This is the output of the optimization and is useful for
%                   determining whether there was a problem during the optimization
% So                The optimal spot leave times
% spotPlaceOpt      Sequence of optimal ready at spot times
% deltaChunk        A matrix of separation times between all pairs of aircraft
%                   in this bin

function [t, place, solution, So, spotPlaceOpt, deltaChunk] = DepMILP(depDataChunk, numFlights, lastFlight, lastAtype, binNum, separationT, spotPlaceChunk)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Initialize variables %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

spotT = depDataChunk(:,1); % Spot ready times of the chunk of data
unimp = depDataChunk(:,2); % Unimpeded taxi times of the chunk of data
alpha = depDataChunk(:,3); % Runway ready times of the chunk of data
aType = depDataChunk(:,4); % Aircraft types of the chunk of data

% Set up a matrix of separation times between all pairs of aircrafts in the
% chunk of data
deltaChunk = zeros(numFlights);
for i = 1 : numFlights
    for j = 1 : numFlights
        if (aType(i) > 0 && aType(j) > 0 && i ~= j)
            deltaChunk(i,j) = separationT(aType(i), aType(j));
        end
    end
end

M = 36000; % A large number, used for CPS and separation times

% z(i,j) is a binary variable that is equal to one if i is before j in the
% sequence
z1 = binvar(numFlights, numFlights, 'full'); % Optimal runway sequence
z2 = binvar(numFlights, numFlights, 'full'); % Optimal spot sequence

% Optimal time assigned for take off
t = sdpvar(numFlights,1);
% Optimal time to leave the spot 
So = t - unimp;

% Determines what place each aircraft takes off in line
sumZ1 = sum(z1');
place = numFlights - sumZ1';
% Determines what place each aircraft leaves the spot in line
sumZ2 = sum(z2');
spotPlaceOpt = numFlights - sumZ2';

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% OBJECTIVES %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Objective is to minimize the maximum takeoff time
% Objective = max(t); % Uncomment this to use minimize makespan

% Objective is to minimize system delay (same result as average delay)
sumT = sum(t);
sumReadyT = sum(alpha);

Objective = (sumT - sumReadyT);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% CONSTRAINTS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% An aircraft cannot take off before it is ready
C1 = [t >= alpha];

% An aircraft cannot leave the spot before it is ready
C2 = [So >= spotT];

% If i is before j then j is not before i
C3 = [z1 + z1' == ones(numFlights) - eye(numFlights)];
C4 = [z2 + z2' == ones(numFlights) - eye(numFlights)];

% Separation times at the runway based on aircraft type are not violated
C5 = [];
for i = 1: numFlights
    for j = 1: numFlights
        if (i~=j)
            C5 = C5 + [t(j)-t(i) >= deltaChunk(i,j)*z1(i,j) - (1 - z1(i,j))*M];
        end
    end
end

% Aircraft leave the spot atleast 1 second apart
C6 = [];
for i = 1: numFlights
    for j = 1: numFlights
        if (i~=j)
            C6 = C6 + [So(j)-So(i) >= 1*z2(i,j)- (1 - z2(i,j))*M];
        end
    end
end

% Separation times between bins are not violated
C7 = [];
if binNum >= 2
    for i = 1: numFlights
        C7 = C7 + [t(i)-lastFlight >= separationT(lastAtype,aType(i))];
    end
end

%%%% CPS CONSTRAINT %%%%
% Aircraft do not move more than numCPS positions from their original spot
% sequence
C8 = [];
numCPS = 0; % Set to M if no CPS constraint is desired
for i = 1: numFlights
    C8 = C8 + [abs(spotPlaceOpt(i)- spotPlaceChunk(i)) <= numCPS];
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Solve %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

Constraints = [C1, C2, C3, C4, C5, C6, C7, C8]; 

% options = sdpsettings('verbose', 0); % Uncomment this line to use Yalmip as the
options = sdpsettings('solver', 'gurobi', 'verbose', 0); % Uncomment this line to use gurobi as a solver

% Solves the optimization using the objection function, the constraints, and the options stated above
solution = solvesdp(Constraints, Objective, options); 

if solution.problem ~= 0 % If the solver has any problems then stop the code
    return
end
