% Created by Julia Baum on September 19, 2011
% 
% This code optimizes the first iteration with the mean fuzzy unimpeded times to get
% desired spot release times. Taxi times are then perturbed and FCFS is
% enforced at the runway. It calls the function DepBinning to perform the
% optimization and FuzzySetup___ and DepartureData___ to set up the data
% that is used.
% 
% INPUTS: 
% None
%
% Change FuzzySetup___ and DepartureData___ (and AT__ if using even
% distribution of aircraft types) so that their endings match. This
% determines the day of data that is used.
%
% OUTPUTS:
% realDelay         Sytem delay of real world takeoff times minus simulated
%                   runway ready times
% dayDelay          Optimal system delay
% delayFCFS         FCFS at runway system delay (non-optimization)
%                   time
% yalmipTime        how long it took Yalmip to set up the MILP
% solverTime        how long it took the solver to solve the MILP]
%
% Currently the outputs for the deterministic world are the system delay
% after optimizing (dayDelay) and the system delay for simulated FCFS (delayFCFS).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Initialize variables %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
numUncert = 500;            % Number of iterations
DepartureData618;           % Departure data from DFW
[theXs, theBs, fuzzyTimes] = FuzzySetup618(numUncert);  % Models unimpeded time perturbations
avgUnimp = [];              % Average over numUncert new unimpeded times
ticTot = tic;               % Start timer that times program
EvenATDist;                 % Even distribution of aircraft types for each day

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Evaluation of deterministic optimization under stochasticity %%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% When k = 1 the unimpeded times are taken from the mean of the triangular
% distribution. The sequence is optimized and system delay is given.
% k=1 represents the the deterministic model. When k > 1
% unimpeded times are randomly chosen from a triangular distribution that
% depends on spots and aircraft types. These iterations represent the
% stochastic model.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Deterministic world %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
int32(depData);
depData(:,2) = theXs*theBs(:,3)*60+90; % Mean triangular taxi times
depData(:,2) = int32(depData(:,2)); % Make the taxi times be integers
depData(:,3) = depData(:,1) + depData(:,2); % New runway ready times (alphas)
% depData(:,4) = AT14; % AT__ gives an even distribution for the day. __ stands for 14, 15, 16 or 18
depData = sortrows(depData,3);
[A, B, D, E, F, G, H, I, J, detFCFSt, delayFCFS, optDelay, solution, yalmipTime, solverTime] = DepBinning(depData); % optimize the sequence to minimize system delay
if solution.problem ~= 0 % If any of the bins have a solver problem then stop the code
    return
end
G = int32(G); % Make the runway sequence be all integers
H = int32(H); % Make the spot sequence be all integers
I = int32(I); % Make the optimal spot sequence be all integers
avgUnimp = [avgUnimp; sum(depData(:,2))/size(depData,1)];
realDelay = sum(sort(realT) - sort(E),1) % Calculate system delay for real world take off times from DFW and simulated runway ready times
optDelay % System delay for optimization
delayFCFS % System delay for non-optimization (FCFS at the runway)
time = toc(ticTot) % end timer and print out timing result

% Find original spot sequence (from spot leave times in the DFW data)
orgSeq = zeros(size(depData,1),1);
for i = 1 : size(depData)
    sortedData = sortrows(depData,3);
    [~, orgPos] = sort(B);
    orgSeq(i,1) = find(orgPos(:,1) == i);
end
% Find optimal spot sequence from optimal spot leave times
optSeq = zeros(size(depData,1),1);
for i = 1 : size(depData)
    [~, optPos] = sort(J);
    optSeq(i,1) = find(optPos(:,1) == i);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Stochastic world %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%
ticStoch = tic;

allData = [A, B, D, E, F, G, H, I, J]; % Gather all data together
sortedAllData = sortrows(allData,-2); % Sort all data into original order (descending by original spot leave times)
% Split the matrix of data into individual vectors again
A = sortedAllData(:,1);
B = sortedAllData(:,2);
D = sortedAllData(:,3);
E = double(sortedAllData(:,4));
F = sortedAllData(:,5);
G = sortedAllData(:,6);
H = sortedAllData(:,7);
I = sortedAllData(:,8);
J = sortedAllData(:,9);

% Perform stochastic iterations
StochasticityB(A, B, D, E, F, G, J, fuzzyTimes, numUncert);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Outputs/ Print outs %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
time = toc(ticStoch)
% Check CPS constraint and check that the flights are not leaving before
% they are ready. checkCPS should be equal to numCPS and ready should be
% equal to zero (or any positive number)
checkCPS = max(abs(optSeq-orgSeq));
for i = 0 : checkCPS
    [i, size(find(abs(optSeq-orgSeq)== i),1)]
end
ready = min(int32(F)-int32(E)); % Check minimum separation times

yalmipTime % how long it took Yalmip to set up the MILP
solverTime % how long it took the solver to solve the MILP



% Problems with optimization
% -7 Solver does not return error codes
% -6 Search space not bounded (bound all variables)
% -5 License problems in solver
% -4 Solver not applicable
% -3 Solver not found
% -2 No suitable solver
% -1 Unknown error
% 0 No problems detected
% 1 Infeasible problem
% 2 Unbounded objective function
% 3 Maximum #iterations or time-limit exceeded
% 4 Numerical problems
% 5 Lack of progress
% 6 Initial solution infeasible
% 7 YALMIP sent incorrect input to solver
% 8 Feasibility cannot be determined
% 9 Unknown problem in solver
% 10 bigM failed (obsolete)
% 11 Other identified error
% 12 Infeasible or unbounded
% 13 YALMIP cannot determine status in solver
% 14 Convexity check failed.
% 15 Problem either infeasible or unbounded

