% Created by Julia Baum on September 19, 2011
%
%
% INPUTS:
% A aircraft types 
% B original spot leave times 
% D original unimpeded times
% E original runway ready times
% F optimal takeoff times
% G optimal takeoff sequence
% J optimal spot leave times
% All of these inputs are for the whole day and should be in a column
% vector.
% fuzzyTimes stochastic unimpeded times
% numUncert number of iterations (number of taxi times for each aircraft)
%
% OUTPUTS:
% none
%
% Currently the outputs for the stochastic world are the percentiles of the numUncert
% iterations for system delay using optimal spot leave times (persOpt) and the system
% delay using the original spot leave times (pers).
% percentShift shows the amount of flights which shifted x positions for
% each iteration. This represents how much the optimal sequence is shifting

function [] = StochasticityB(A, B, D, E, F, G, J, fuzzyTimes, numUncert)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Initialize variables %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

delays = zeros(numUncert,1);              % System delay when not optimizing
delayOpts = zeros(numUncert,1);           % System delay when optimizing
newTaxiTimes = zeros(size(A,1));          % New taxi times created by triangular distributions for numUncert iterations
changes = zeros(numUncert,1);             % Average perturbation of the taxi times
avgNewUnimp = zeros(numUncert,1);         % Average new unimpeded times
avgOrgUnimp = zeros(numUncert,1);         % Average original unimpeded times              
maxChanges = zeros(numUncert,1);          % Maximum amount that a taxi time changed
ticTot = tic;                             % Start timer that times program
seqTest1Total = zeros(numUncert,1);       % The number of total positions are shifted in the sequence
seqTest2Total = zeros(numUncert,1);       % The number of pairs switch in the sequence relative to each other
seqTest3Total = zeros(numUncert,1);       % seqTest2 with added weights to show how far apart they shifted
sepTest1Total = zeros(numUncert,1);       % The number of aircraft that violated separation times by arriving at the runway too early
sepTest2Total = zeros(numUncert,1);       % The average amount of time that the aircraft in sepTest1 arrive too early by
avgChange = zeros(numUncert,1);           % Average change of the taxi times
maxChange = zeros(numUncert,1);           % Maximum change of the taxi times
minChange = zeros(numUncert,1);           % Minimum change of the taxi times

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Stochastic Model with uncertain taxi times %%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Unimpeded times are randomly chosen from a triangular distribution that
% depends on spots and aircraft types. These k iterations represent the
% stochastic model.

for k = 1 : numUncert
    G = int32(G);
    
    % Original data
    depData = [B, D, E, A];
    
    % Optimized data, with new unimpeded and alpha runway ready times
    tempOptDepData = depData;
    tempOptDepData(:,1) = J;
    tempOptDepData(:,2) = int32(fuzzyTimes(:,k));
    tempOptDepData(:,3) = tempOptDepData(:,1) + tempOptDepData(:,2);
    
    % Record the average unimpeded time
    avgOrgUnimp(k) = sum(depData(:,2))/size(depData,1);
    
    % Determine new runway sequence due to new taxi times
    newRunwaySeq = zeros(size(tempOptDepData,1),1);
    % Sort by alpha runway ready times, get indexes
    [~, FCFSpos] = sort(tempOptDepData(:,3));
    for i = 1 : size(FCFSpos)
        % Save new runway index for aircraft i
        newRunwaySeq(FCFSpos(i),1) = i;
    end
    newRunwaySeq = int32(newRunwaySeq);
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%% FCFS takeoff times with ORIGINAL alphas %%%%
    
    % Update non-optimized data with new taxi times and runway ready times
    nonOptDepData = depData;
    nonOptDepData(:,2) = fuzzyTimes(:,k);
    nonOptDepData(:,3) = nonOptDepData(:,1) + nonOptDepData(:,2);
    
    % Sort data by the new alphas
    nonOptDepData = sortrows(nonOptDepData,3);
    
    % Separation matrix: leading rows, trailing columns
    separationT = [60, 60, 60, 60;
        90, 60, 60, 60;
        120, 120, 90, 120;
        120, 90, 90, 90];
    
    % Vector of new alpha runway ready times
    FCFSt = newRunwayTimes(nonOptDepData, separationT);
    
    %%%%                                         %%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%% FCFS takeoff times with OPTIMAL alphas  %%%%
    
    % Sort data by the alphas
    optDepData = sortrows(tempOptDepData,3);
    
    % Vector of new alpha runway ready times
    [optFCFSt, alphaOpt, deltaOpt] = newRunwayTimes(optDepData, separationT);
    
    %%%%                                         %%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    % System delay due to perturbations (optimized and not optimized)
    delay = sum(FCFSt - E); % Delay when using original spot leave times
    delayOpt = sum(optFCFSt - E); % Delay when using optimal spot leave times
    delays(k) = delay;
    delayOpts(k) = delayOpt;
    
    % Vector of changes in unimpeded times for one iteration
    unimpChange = tempOptDepData(:,2) - depData(:,2);
    
    % Average change in unimpeded time for one iteration
    avgChange = sum(abs(unimpChange))/size(depData,1);
    
    % Most that one unimpeded time changed for this iteration
    maxChange = max(unimpChange);
    minChange = min(unimpChange);
    
    % Vectors of above values (except last one)
    newTaxiTimes(:,k) = unimpChange;
    changes(k) = avgChange;
    maxChanges(k) = maxChange;
    minChanges(k) = minChange;
    avgNewUnimp(k) = sum(tempOptDepData(:,2))/size(depData,1);   
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%% Sequence Analysis %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    % G is the optimal sequence at the runway
    % newRunwaySeq is the sequence at the runway after perturbations
    seqTest1Total(k) = sum(abs(newRunwaySeq-G)); % Total number of shifts
    
    x = [G, newRunwaySeq];
    x2 = sortrows(x,2);
    
    seqTest2 = 0;
    seqTest3 = 0;
    for i = 1 : size(depData,1)-1
        iPosOpt = x2(i,1);
        for j = i+1 : size(depData,1)
            if x2(j,1) < iPosOpt
                seqTest2 = seqTest2 + 1;
                seqTest3 = seqTest3 + iPosOpt - x2(j,1);
            end
        end
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%% Separation Time Analysis %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    sepTest1 = 0;
    sepTest2 = 0;
    for i = 1 : size(alphaOpt,1)-1
        if alphaOpt(i+1) - alphaOpt(i) < deltaOpt(i, i+1)
            sepTest1 = sepTest1 + 1;
            sepTest2 = sepTest2 + deltaOpt(i, i+1) - (alphaOpt(i+1)-alphaOpt(i));
        end
    end 

    seqTest2Total(k) = seqTest2; % Number of pairwise swaps
    seqTest3Total(k) = seqTest3; % Number of pairwise swaps weighted by how many positions they shifted
    sepTest1Total(k) = sepTest1; % Number of aircraft that arrived too early violating separation requirements
    sepTest2Total(k) = sepTest2; % Amount of time that the aircraft arrived too early by (in seconds)
end

time = toc(ticTot) % End timer

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Outputs/ Print outs %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

avgSeqTest1 = sum(seqTest1Total)/numUncert % Average total number of shifts
avgSeqTest2 = sum(seqTest2Total)/numUncert % Average number of pairwise swaps
avgSeqTest3 = sum(seqTest3Total)/numUncert % Average number of pairwise swaps weighted by how many positions they shifted
avgSepTest1 = sum(sepTest1Total)/numUncert % Average number of aircraft that arrived too early violating separation requirements
avgSepTest2 = sum(sepTest2Total)/numUncert % Average total amount of time that the aircraft arrived too early by (in seconds)

avgChanges = sum(changes)/numUncert % Average change of taxi times due to uncertainties
maxChanges = max(maxChanges) % Maximum change of taxi times
minChanges = min(minChanges) % Minimum change of taxi times

% Delay percentiles
avgDelayOpts = (delayOpts/size(depData,1))';
persOpt = prctile(avgDelayOpts, [5 25 50 75 95 100])';
avgDelays = (delays/size(depData,1))';
pers = prctile(avgDelays, [5 25 50 75 95 100])';

% Stochastic departure delay
delayOptStoch = persOpt(3)*size(depData,1) 
delayStoch = pers(3)*size(depData, 1)
